/************************CASP_COPYRIGHT**********************************
** Copyright (c) AadhunikLabs. All rights reserved.
** This file is part of CASP.
** Licensees holding valid licenses may use this file in accordance
** with the license agreement provided with the software.
*************************************************************************/
#include "CustomBlockSelfBalancing720378270.h"
#include "simcode.h"
#include "math.h"
#include "sim_helper_fg.h"

CustomBlockSelfBalancing720378270::CustomBlockSelfBalancing720378270(){
}

CustomBlockSelfBalancing720378270::~CustomBlockSelfBalancing720378270(){
}

void CustomBlockSelfBalancing720378270::Initialize(const volatile BLOCK_CONSTRUCTOR_PARAM_STRUCT *arg){
    m_prerun_count = 0;
}

void CustomBlockSelfBalancing720378270::PreRun(const volatile GLOBAL_PRERUN_PARAM_STRUCT *arg, _FXDPT_FLOAT *pid_in0, _FXDPT_FLOAT *pid_in1, _INT32 *estop, _BYTE *data_in, _INT32 *comm_len, _INT32 *left_out0, _INT32 *left_out1, _INT32 *right_out0, _INT32 *right_out1, _FXDPT_FLOAT *servo0, _FXDPT_FLOAT *servo1, _INT32 *led0, _FXDPT_FLOAT *sen_offset, _FXDPT_FLOAT *sp_l, _FXDPT_FLOAT *sp_r, _INT32 *rst_i){
    if(m_prerun_count++ != 1)
    {
        *pid_in0 = 0;
        *pid_in1 = 0;
        for(int i=0; i<20; ++i)
            data_in[i] = 0;
        *comm_len = 0;
        for(int i=0; i<3; ++i)
            estop[i] = 0;
        return;
    }
    m_left_dir_prev = 0;
    m_right_dir_prev = 0;
    m_left_speed_prev = 0;
    m_right_speed_prev = 0;
    //
    *servo0 = *((_FXDPT_FLOAT*)&data_in[20]);
    *servo1 = *((_FXDPT_FLOAT*)&data_in[24]);
    *led0 = data_in[28];
    //
    *left_out0 = *left_out1 = *right_out0 = *right_out1 = 0;
    *rst_i = 1000;
    *sp_l = *sp_r = 0;
    m_comm_tick_prev = m_speed_tick_prev = arg->rt_clk_tick;
    m_usr_offset = 0;
}

void CustomBlockSelfBalancing720378270::Run(const volatile GLOBAL_RUN_PARAM_STRUCT *arg, _FXDPT_FLOAT *pid_in0, _FXDPT_FLOAT *pid_in1, _INT32 *estop, _BYTE *data_in, _INT32 *comm_len, _INT32 *left_out0, _INT32 *left_out1, _INT32 *right_out0, _INT32 *right_out1, _FXDPT_FLOAT *servo0, _FXDPT_FLOAT *servo1, _INT32 *led0, _FXDPT_FLOAT *sen_offset, _FXDPT_FLOAT *sp_l, _FXDPT_FLOAT *sp_r, _INT32 *rst_i){
    //Note: run this block in sync with PID blocks otherwise PID integral reset signal may not work
    *servo0 = *((_FXDPT_FLOAT*)&data_in[20]);
    *servo1 = *((_FXDPT_FLOAT*)&data_in[24]);
    *led0 = data_in[28];
    //adjust set point angle w.r.t manual control
    _BIT comm_stopped = false;
    if(*comm_len > 0)
        m_comm_tick_prev = arg->rt_clk_tick;
    else
    {
        if(arg->rt_clk_tick - m_comm_tick_prev > 500000)
            comm_stopped = true;
    }
    //sensor offsets
    m_usr_offset += ((_FLOAT)*((_INT32*)&data_in[16])) * 0.01;
    _FLOAT offset = *sen_offset + m_usr_offset;
    //
    _FLOAT &spl = *sp_l;
    _FLOAT &spr = *sp_r;
    _INT32 left_dir = *((_INT32*)&data_in[0]);
    _INT32 left_speed = *((_INT32*)&data_in[4]);
    _INT32 right_dir = *((_INT32*)&data_in[8]);
    _INT32 right_speed = *((_INT32*)&data_in[12]);
    spl = spr = 0;
    //initially, speed is required to overcome static friction. reduce the speed after some time for stability
    if(left_speed != 0 || right_speed != 0)
    {
        if(m_left_speed_prev == 0 && m_right_speed_prev == 0)//user just given command to move or rotate
            m_speed_tick_prev = arg->rt_clk_tick;
        if(arg->rt_clk_tick - m_speed_tick_prev > 900000)
        {
            left_speed /= 4;
            right_speed /= 4;
        }
        else if(arg->rt_clk_tick - m_speed_tick_prev > 500000)
        {
            left_speed /= 2;
            right_speed /= 2;
        }
        else if(arg->rt_clk_tick - m_speed_tick_prev > 300000)
        {
            left_speed = left_speed*2/3;
            right_speed = right_speed*2/3;
        }
    }
    //reset manual control when user is not connected
    if(comm_stopped)
    {
        spl = 0;
        spr = 0;
        //*servo0 = *servo1 = 0;//we dont know offsets and it is not mandatory to make is zero
        *led0 = 0;
    }
    else
    {
        spl = left_speed;
        spr = right_speed;
        //set direction
        spl = (left_dir)?spl:-spl;
        spr = (right_dir)?spr:-spr;
        //to actual
        spl /= 19.0;
        spr /= 19.0;
    }
    spl += offset;
    spr += offset;
    //reset pid block integral term on following logics
    *rst_i = 0;
    if(left_speed == 0 && right_speed == 0)//only during braking
    {
        if(left_dir != m_left_dir_prev || right_dir != m_right_dir_prev)//when change in direction
            *rst_i = 10;
        if(left_speed != m_left_speed_prev || right_speed != m_right_speed_prev)
            *rst_i = 10;
    }
    *rst_i -= 1;
    //
    m_left_dir_prev = left_dir;
    m_right_dir_prev = right_dir;
    m_left_speed_prev = left_speed;
    m_right_speed_prev = right_speed;

    //process outputs
    _FLOAT in[2], out[2];
    in[0] = -*pid_in0;
    in[1] = -*pid_in1;
    for(int i=0; i<2; ++i)
    {
        out[i] = in[i];
        //absolute
        if(out[i] < 0) out[i] = -out[i];
        //map to minimum & maximum pwm outputs to turn the wheel
        out[i] = MapRangeT(out[i], (_FLOAT)pwm_max_limit, (_FLOAT)pwm_min_offset, (_FLOAT)100, (_FLOAT)0);
        //stop when the input angle is greater than set angle
        if(estop[0] > em_stop_value || estop[0] < -em_stop_value)
            out[i] = 0;
    }
    //set motor speed and direction
    if(in[0] < 0)
    {
        *left_out0 = out[0];
        *left_out1 = 0;
    }
    else
    {
        *left_out0 = 0;
        *left_out1 = out[0];
    }
    if(in[1] < 0)
    {
        *right_out0 = 0;
        *right_out1 = out[1];
    }
    else
    {
        *right_out0 = out[1];
        *right_out1 = 0;
    }
}

void CustomBlockSelfBalancing720378270::PostRun(const volatile GLOBAL_POSTRUN_PARAM_STRUCT *arg, _FXDPT_FLOAT *pid_in0, _FXDPT_FLOAT *pid_in1, _INT32 *estop, _BYTE *data_in, _INT32 *comm_len, _INT32 *left_out0, _INT32 *left_out1, _INT32 *right_out0, _INT32 *right_out1, _FXDPT_FLOAT *servo0, _FXDPT_FLOAT *servo1, _INT32 *led0, _FXDPT_FLOAT *sen_offset, _FXDPT_FLOAT *sp_l, _FXDPT_FLOAT *sp_r, _INT32 *rst_i){
}
